<?php

declare(strict_types=1);

namespace Mediastrategi\PostNord\Helper;

use Magento\Framework\Api\FilterBuilder;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\Exception\LocalizedException;
use Mediastrategi\PostNord\Api\Data\TransactionInterfaceFactory;
use Mediastrategi\PostNord\Api\TransactionRepositoryInterface;
use Mediastrategi\PostNord\Api\Data\TransactionInterface;

class ApiTransaction extends \Magento\Framework\App\Helper\AbstractHelper
{
    /**
     * @var TransactionRepositoryInterface
     */
    private $transactionRepository;

    /**
     * @var TransactionInterfaceFactory
     */
    private $transactionInterfaceFactory;

    /**
     * @var SearchCriteriaBuilder
     */
    private $searchCriteriaBuilder;

    /**
     * @var FilterBuilder
     */
    private $filterBuilder;

    public function __construct(
        \Magento\Framework\App\Helper\Context $context,
        TransactionRepositoryInterface $transactionRepository,
        TransactionInterfaceFactory $transactionInterfaceFactory,
        SearchCriteriaBuilder $searchCriteriaBuilder,
        FilterBuilder $filterBuilder
    ) {
        parent::__construct($context);
        $this->transactionRepository       = $transactionRepository;
        $this->transactionInterfaceFactory = $transactionInterfaceFactory;
        $this->searchCriteriaBuilder       = $searchCriteriaBuilder;
        $this->filterBuilder               = $filterBuilder;
    }

    public function clean()
    {
        $filter = $this->filterBuilder
            ->setField(TransactionInterface::ADDED)
                ->setConditionType('lteq')
                ->setValue(date('Y-m-d H:i:s', strtotime('-14 days')))
                ->create();

        $this->searchCriteriaBuilder->addFilters([$filter]);

        $searchCriteria = $this->searchCriteriaBuilder->create();
        try {
            $transactionsToDelete = $this->transactionRepository->getList($searchCriteria)->getItems();
        } catch (LocalizedException $e) {
            $this->_logger->error('Could not get Transaction List');

            return; // Nothing left to do
        }

        foreach ($transactionsToDelete as $transaction) {
            try {
                $this->transactionRepository->delete($transaction);
            } catch (LocalizedException $e) {
                $this->_logger->error(
                    'Could not delete Transaction (ID: ' . $transaction->getTransactionId() . '): ' . $e->getMessage()
                );
            }
        }
    }

    public function add(array $data): ?int
    {
        $newTransaction = $this->transactionInterfaceFactory->create();
        if (isset($data[TransactionInterface::ORDER_IDS])) {
            $newTransaction->setOrderIds($data[TransactionInterface::ORDER_IDS]);
        }
        if (isset($data[TransactionInterface::QUOTE_ID])) {
            $newTransaction->setQuoteId($data[TransactionInterface::QUOTE_ID]);
        }
        $newTransaction->setRequestMethod($data[TransactionInterface::REQUEST_METHOD]);
        $newTransaction->setRequestUri($data[TransactionInterface::REQUEST_URI]);
        $newTransaction->setRequestHeaders(
            $this->maskSensitiveData(
                $data[TransactionInterface::REQUEST_HEADERS]
            )
        );
        $newTransaction->setRequestBody($data[TransactionInterface::REQUEST_BODY]);
        $newTransaction->setResponseStatusCode($data[TransactionInterface::RESPONSE_STATUS_CODE]);
        $newTransaction->setResponseBody($data[TransactionInterface::RESPONSE_BODY]);
        $newTransaction->setResponseHeaders($data[TransactionInterface::RESPONSE_HEADERS]);
        if (isset($data[TransactionInterface::SHIPMENT_IDS])) {
            $newTransaction->setShipmentIds($data[TransactionInterface::SHIPMENT_IDS]);
        }

        try {
            $model = $this->transactionRepository->save($newTransaction);
            return $model->getTransactionId();
        } catch (LocalizedException $e) {
            $this->_logger->error(
                'Could not save Transaction: ' . $e->getMessage() . ', Data: ' . print_r($data, true)
            );
        }
        return null;
    }

    private function maskSensitiveData(string $data): string
    {
        static $sensitiveKeys = [
            'authorization',
            'x-bap-key',
        ];

        try {
            $decodedData = json_decode(
                $data,
                true
            );
        } catch (\Exception $e) {
            $decodedData = [];
        }
        if (!empty($decodedData)) {
            $updated = 0;
            foreach ($sensitiveKeys as $sensitiveKey) {
                if (isset($decodedData[$sensitiveKey])) {
                    unset($decodedData[$sensitiveKey]);
                    $updated++;
                }
            }
            if ($updated) {
                return json_encode(
                    $decodedData
                );
            }
        }
        return $data;
    }
}
