<?php

declare(strict_types=1);

namespace Mediastrategi\PostNord\Helper;

class Cache extends \Magento\Framework\App\Helper\AbstractHelper
{
    /**
     * @var \Mediastrategi\PostNord\Model\Cache\Api
     */
    private $apiCacheModel;

    /**
     * @var \Magento\Store\Model\StoreManagerInterface
     */
    private $storeManager;

    public function __construct(
        \Mediastrategi\PostNord\Model\Cache\Api $apiCacheModel,
        \Magento\Framework\App\Helper\Context $context,
        \Magento\Store\Model\StoreManagerInterface $storeManager
    ) {
        parent::__construct($context);
        $this->apiCacheModel = $apiCacheModel;
        $this->storeManager = $storeManager;
    }

    public function saveApiToken(
        string $apiIdentifier,
        string $apiKey,
        string $apiToken
    ): void {
        $key = $this->getApiTokenCacheKey(
            $apiIdentifier,
            $apiKey
        );
        $this->apiCacheModel->save(
            $apiToken,
            $key,
            [\Mediastrategi\PostNord\Model\Cache\Api::CACHE_TAG],
            60 * 60 * 24 // 24 hours
        );
    }

    public function loadApiToken(
        string $apiIdentifier,
        string $apiKey
    ): string {
        $key = $this->getApiTokenCacheKey(
            $apiIdentifier,
            $apiKey
        );
        if (
            $data = $this->apiCacheModel->load($key)
        ) {
            return (string) $data;
        }
        return '';
    }

    public function saveQuery(
        array $arguments,
        \Mediastrategi\PostNordAPI\Transaction $transaction
    ): void {
        $key = $this->getQueryCacheKey(
            $arguments
        );
        $data =
            $this->getEncodedTransaction($transaction);
        $this->apiCacheModel->save(
            $data,
            $key,
            [\Mediastrategi\PostNord\Model\Cache\Api::CACHE_TAG],
            60 * 30 // 30 minutes
        );
    }

    public function loadQuery(
        array $arguments
    ): ?\Mediastrategi\PostNordAPI\Transaction {
        $key = $this->getQueryCacheKey(
            $arguments
        );
        if (
            $data = $this->apiCacheModel->load($key)
        ) {
            try {
                $decodedData = json_decode(
                    $data,
                    true
                );
                // Validate cached data
                if (
                    !empty($decodedData['request_method'])
                    && is_string($decodedData['request_method'])
                    && !empty($decodedData['request_uri'])
                    && is_string($decodedData['request_uri'])
                    && isset($decodedData['request_headers'])
                    && is_array($decodedData['request_headers'])
                    && isset($decodedData['request_body'])
                    && is_string($decodedData['request_body'])
                    && !empty($decodedData['response_status_code'])
                    && is_int($decodedData['response_status_code'])
                    && isset($decodedData['response_body'])
                    && is_string($decodedData['response_body'])
                    && isset($decodedData['response_headers'])
                    && is_array($decodedData['response_headers'])
                ) {
                    return new \Mediastrategi\PostNordAPI\Transaction(
                        $decodedData['request_method'],
                        $decodedData['request_uri'],
                        $decodedData['request_headers'],
                        $decodedData['request_body'],
                        $decodedData['response_status_code'],
                        $decodedData['response_headers'],
                        $decodedData['response_body']
                    );
                }
            } catch (\Exception $e) {
                $decodedData = null;
            }
        }
        return null;
    }

    public function saveLicenseValidationResponse(
        array $license,
        array $licenseValidationResponse
    ): void {
        $key = $this->getLicenseValidationResponseCacheKey(
            $license
        );
        $data =
            json_encode($licenseValidationResponse);
        $this->apiCacheModel->save(
            $data,
            $key,
            [\Mediastrategi\PostNord\Model\Cache\Api::CACHE_TAG],
            21600 // 6 hours
        );
    }

    public function loadLicenseValidationResponse($license): ?array
    {
        $key = $this->getLicenseValidationResponseCacheKey(
            $license
        );
        if (
            $data = $this->apiCacheModel->load($key)
        ) {
            try {
                $decodedData = json_decode(
                    $data,
                    true
                );
                // Validate cached data
                if (
                    isset($decodedData['valid'])
                ) {
                    return $decodedData;
                }
            } catch (\Exception $e) {
                $decodedData = null;
            }
        }
        return null;
    }

    private function getApiTokenCacheKey(
        string $apiIdentifier,
        string $apiKey
    ): string {
        return sprintf(
            'api_token_%s_%s',
            $apiIdentifier,
            $apiKey
        );
    }

    private function getQueryCacheKey(
        array $arguments
    ): string {
        $apiIdentifier = $this->getConfig(
            'carriers/mspostnord/api/identifier'
        );
        $apiKey = $this->getConfig(
            'carriers/mspostnord/api/key'
        );
        $limitServicePoints = (int) $this->getConfig(
            'carriers/mspostnord/limit_service_points'
        );
        return sprintf(
            'query_%s_%s_%s_%s',
            $apiIdentifier,
            $apiKey,
            $limitServicePoints,
            $this->getFlattenedArray(
                $arguments
            )
        );
    }

    private function getLicenseValidationResponseCacheKey(
        array $arguments
    ): string {
        return sprintf(
            'license_%s',
            $this->getFlattenedArray(
                $arguments
            )
        );
    }

    private function getEncodedTransaction(
        \Mediastrategi\PostNordAPI\Transaction $transaction
    ): string {
        $jsonDecoded = [
            'request_method' =>
                $transaction->getRequestMethod(),
            'request_uri' =>
                $transaction->getRequestUri(),
            'request_headers' => $this->maskSensitiveData(
                $transaction->getRequestHeaders()
            ),
            'request_body' =>
                $transaction->getRequestBody(),
            'response_status_code' =>
                $transaction->getResponseStatusCode(),
            'response_headers' =>
                $transaction->getResponseHeaders(),
            'response_body' =>
                $transaction->getResponseBody(),
        ];
        return json_encode($jsonDecoded);
    }

    private function maskSensitiveData(array $data): array
    {
        static $sensitiveKeys = [
            'authorization',
            'x-bap-key',
        ];
        foreach ($sensitiveKeys as $sensitiveKey) {
            if (isset($data[$sensitiveKey])) {
                unset($data[$sensitiveKey]);
            }
        }
        return $data;
    }

    private function getFlattenedArray(array $data): string
    {
        $string = '';
        $i = 0;
        foreach ($data as $key => $value) {
            $key = (string) $key;
            if ($i) {
                $string .= '&';
            }
            if (is_array($value)) {
                $value = $this->getFlattenedArray($value);
                $string .= $value;
                $i++;
                continue;
            } elseif (is_object($value)) {
                $value = (array) $value;
                $value = $this->getFlattenedArray($value);
                $string .= $value;
                $i++;
                continue;
            } elseif (is_resource($value)) {
                $value = '-';
            } else {
                $value = (string) $value;
            }
            $string .= "$key=$value";
            $i++;
        }
        return $string;
    }

    /**
     * NOTE We cannot use this method from Meta because
     * that would create a circular dependency
     */
    private function getConfig(
        string $path,
        int $storeId = null
    ) {
        try {
            $value = $this->scopeConfig->getValue(
                $path,
                \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
                $storeId ?? $this->storeManager->getStore()->getId()
            );
            return $value;
        } catch (\Exception $e) {
            return '';
        }
    }
}
