<?php

declare(strict_types=1);

namespace Mediastrategi\PostNord\Helper;

use Mediastrategi\PostNord\Api\Data\TransactionInterface;

class Meta extends \Magento\Framework\App\Helper\AbstractHelper
{
    public const ABOUT_URL = 'https://www.oktagon.se/';
    public const ABOUT_EMAIL = 'support@oktagon.se';
    public const ABOUT_PHONE = '(+46) 021-470 88 30';
    public const ABOUT_LOGO = 'https://license.oktagon.se/images/m2-postnord.png';
    public const ORDER_PROCESSING_BOOK = 'book';
    public const ORDER_PROCESSING_RESERVE = 'reserve';

    /**
     * @var \Magento\Store\Model\StoreManagerInterface
     */
    private $storeManager;

    /**
     * @var \Magento\Framework\App\ProductMetadataInterface
     */
    private $productMetadata;

    /**
     * @var \Magento\Store\Api\Data\StoreInterface
     */
    private $storeInterface;

    /**
     * @var \Magento\Framework\Locale\Resolver
     */
    private $localeResolver;

    /**
     * @var \Magento\Framework\Module\Dir
     */
    private $moduleDir;

    /**
     * @var \Mediastrategi\PostNord\Helper\Api
     */
    private $apiHelper;

    /**
     * @var \Mediastrategi\PostNord\Helper\Cache
     */
    private $cacheHelper;

    /**
     * @var \Magento\Directory\Model\RegionFactory
     */
    private $regionFactory;

    /**
     * @var \Magento\Backend\Model\Auth\Session
     */
    private $authSession;

    /**
     * @var ApiTransaction
     */
    private $apiTransactionHelper;
    /**
     * @var \Magento\Customer\Helper\Session\CurrentCustomer
     */
    private $currentCustomer;
    /**
     * @var \Magento\Tax\Api\TaxCalculationInterface
     */
    private $taxCalculation;

    private $lastShippingOptionsResult = [];

    public function __construct(
        \Magento\Framework\App\Helper\Context $context,
        \Magento\Store\Model\StoreManagerInterface $storeManager,
        \Magento\Framework\App\ProductMetadataInterface $productMetadata,
        \Magento\Store\Api\Data\StoreInterface $storeInterface,
        \Magento\Framework\Locale\Resolver $localeResolver,
        \Magento\Framework\Module\Dir $moduleDir,
        \Mediastrategi\PostNord\Helper\Api $apiHelper,
        \Magento\Directory\Model\RegionFactory $regionFactory,
        \Magento\Backend\Model\Auth\Session $authSession,
        \Mediastrategi\PostNord\Helper\Cache $cacheHelper,
        \Mediastrategi\PostNord\Helper\ApiTransaction $apiTransactionHelper,
        \Magento\Customer\Helper\Session\CurrentCustomer $currentCustomer,
        \Magento\Tax\Api\TaxCalculationInterface $taxCalculation
    ) {
        parent::__construct($context);
        $this->apiHelper            = $apiHelper;
        $this->storeManager         = $storeManager;
        $this->productMetadata      = $productMetadata;
        $this->storeInterface       = $storeInterface;
        $this->localeResolver       = $localeResolver;
        $this->moduleDir            = $moduleDir;
        $this->regionFactory        = $regionFactory;
        $this->authSession          = $authSession;
        $this->cacheHelper          = $cacheHelper;
        $this->apiTransactionHelper = $apiTransactionHelper;
        $this->currentCustomer      = $currentCustomer;
        $this->taxCalculation       = $taxCalculation;
    }

    public function isNewVersionAvailable(): bool
    {
        $currentVersion =
            $this->getExtensionVersion();
        $latestVersion  =
            $this->getLatestVersion();
        return version_compare(
            $currentVersion,
            $latestVersion,
            '<'
        );
    }

    public function getLastShippingOptionsResult(): array
    {
        return $this->lastShippingOptionsResult;
    }

    public function isValidApiCredentials(): int
    {
        $query = [
            'currency'      =>
                $this->storeManager->getStore()->getBaseCurrency()->getCurrencyCode(),
            'language'      =>
                $this->getLanguageCode(),
            'recipient'     => [
                'countryCode' => $this->getConfig(
                    \Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_COUNTRY_ID
                ),
                'postalCode'  => $this->getConfig(
                    \Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_ZIP
                ),
                'type'        =>
                    'personal',
            ],
            'totalPrice'    =>
                25000,
            'weightInGrams' =>
                1000,
        ];
        try {
            $transaction =
                $this->getShippingOptionsForQuery($query);
            if ($transaction === null) {
                return -1;
            } else {
                return 1;
            }
        } catch (\Exception $e) {
            return 0;
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Exception
     */
    public function getShippingOptionsForQuery(
        array $query,
        string $orderIds = null,
        string $shipmentIds = null,
        string $quoteId = null
    ): ?array {
        $loadedFromCache = false;
        $this->log(
            (string) __(
                '%1 %2',
                __METHOD__,
                var_export($query, true)
            )
        );
        if (!$this->hasProperUsage()) {
            return null;
        }
        $transaction = null;
        if (
            $transaction =
                $this->cacheHelper->loadQuery($query)
        ) {
            $loadedFromCache = true;
            $this->log(
                (string) __('Query cache hit')
            );
        } else {
            $this->log(
                (string) __('Query cache miss')
            );
        }
        if (!$transaction) {
            $transaction = $this->getTransactionForQuery(
                $query,
                $orderIds,
                $shipmentIds,
                $quoteId
            );
        }
        if ($transaction === null) {
            // Case were API connection is down
            return null;
        }

        if ($transaction->getResponseStatusCode() === 200) {
            $decoded =
                $transaction->getResponseBodyDecoded();
            if (
                isset($decoded['shipmentOptions'])
                && is_array($decoded['shipmentOptions'])
            ) {
                $shippingOptions =
                    &$decoded['shipmentOptions'];
                $shippingTaxClassId  = $this->scopeConfig->getValue('tax/classes/shipping_tax_class');
                $shippingIncludesTax = (int) $this->scopeConfig->getValue('tax/calculation/shipping_includes_tax');
                $customerId = $this->currentCustomer->getCustomerId();
                $storeId = $this->storeManager->getStore()->getId();
                $vat = $this->taxCalculation->getDefaultCalculatedRate(
                    $shippingTaxClassId,
                    $customerId,
                    $storeId
                );
                $vatDivisor = 1 + ($vat / 100);
                if ($shippingOptions) {
                    $limitServicePoints = (int) $this->getConfig(
                        'carriers/mspostnord/limit_service_points'
                    );
                    foreach ($shippingOptions as &$shippingOption) {
                        // Convert prices to decimals
                        $price = $shippingOption['price'] / 100;
                        if ($shippingIncludesTax === 0) { // Without VAT
                            $priceBeforeVat = $price / $vatDivisor;
                            $shippingOption['price'] = (float) round($priceBeforeVat, 2);
                        } else { // With VAT
                            $shippingOption['price'] = (float) round($price, 2);
                        }

                        // Add a formatted name for each agent
                        if (!empty($shippingOption['servicePoints'])) {
                            foreach ($shippingOption['servicePoints'] as &$servicePoint) {
                                $servicePoint['formattedName'] = $this->getServicePointFormattedName(
                                    $servicePoint
                                );
                            }
                            if (
                                $limitServicePoints
                                && count($shippingOption['servicePoints']) > $limitServicePoints
                            ) {
                                array_splice(
                                    $shippingOption['servicePoints'],
                                    $limitServicePoints
                                );
                            }
                            unset($servicePoint); // Prevent accidental side effects
                        }

                        /**
                         * Set 'deliveryMethod' and 'longDescription' in a way
                         * that is compatible with new API
                         */
                        $shippingOption['deliveryMethod'] = str_replace(
                            " \n",
                            ". ",
                            $shippingOption['deliveryMethod']
                        );
                        if (!isset($shippingOption['longDescription'])) {
                            $shippingOption['longDescription'] = '';
                        }
                        $shippingOption['description'] = str_replace(
                            " \n",
                            '. ',
                            $shippingOption['description']
                        );

                        if (!empty($shippingOption['deliveryEstimate'])) {
                            $shippingOption['deliveryEstimateText'] = $this->getFormattedETA(
                                $shippingOption['deliveryEstimate']
                            );
                        } else {
                            $shippingOption['deliveryEstimateText'] = '';
                        }
                    }
                    unset($shippingOption); // prevent accidental side effects
                }

                if (!$loadedFromCache) {
                    $this->cacheHelper->saveQuery(
                        $query,
                        $transaction
                    );
                    $this->log(
                        (string) __(
                            'Successfully saved transaction-object in cache'
                        )
                    );
                }

                $this->lastShippingOptionsResult =
                    $decoded['shipmentOptions'];
                return $decoded['shipmentOptions'];
            }

            throw new \Exception(
                sprintf(
                    (string) __(
                        'Unexpected response body from server: %s'
                    ),
                    $transaction->getResponseBody()
                )
            );
        } else {
            throw new \Exception(
                sprintf(
                    (string) __(
                        'Unexpected response status-code from server: %s and body %s'
                    ),
                    $transaction->getResponseStatusCode(),
                    $transaction->getResponseBody()
                )
            );
        }
    }

    public function getFormattedETA(
        array $etaData
    ): string {
        $formattedEta = '';
        if (
            isset($etaData['interval'])
            || isset($etaData['earliest'], $etaData['latest'])
        ) {
            $earliestDays = 0;
            $earliestDate = '';
            $latestDays = 0;
            $latestDate = '';
            $isWeekdays = false;
            if (
                isset(
                    $etaData['interval'],
                    $etaData['interval']['earliest'],
                    $etaData['interval']['latest']
                )
                && preg_match(
                    '/^[0-9]+$/',
                    (string) $etaData['interval']['earliest']
                ) === 1
                && preg_match(
                    '/^[0-9]+$/',
                    (string) $etaData['interval']['latest']
                ) === 1
            ) {
                // This estimate is always in weekdays
                $earliestDays =
                    (int) $etaData['interval']['earliest'];
                $latestDays =
                    (int) $etaData['interval']['latest'];
                $isWeekdays = true;
            } elseif (
                isset(
                    $etaData['earliest'],
                    $etaData['latest']
                )
                && preg_match(
                    "/^\d{4}-\d{2}-\d{2}$/",
                    (string) $etaData['earliest']
                ) === 1
                && preg_match(
                    "/^\d{4}-\d{2}-\d{2}$/",
                    (string) $etaData['latest']
                ) === 1
            ) {
                // This estimate can be in non-weekdays
                $earliestDate =
                    (string) $etaData['earliest'];
                $latestDate =
                    (string) $etaData['latest'];
            }

            if ($isWeekdays) {
                if (
                    $earliestDays
                    !== $latestDays
                ) {
                    $formattedEta = sprintf(
                        (string) __('%s-%s weekdays'),
                        $earliestDays,
                        $latestDays
                    );
                } else {
                    if ($earliestDays > 1) {
                        $formattedEta = sprintf(
                            (string) __('%s weekdays'),
                            $earliestDays
                        );
                    } else {
                        $formattedEta = sprintf(
                            (string) __('%s weekday'),
                            $earliestDays
                        );
                    }
                }
            } else {
                if (
                    $earliestDate
                    !== $latestDate
                ) {
                    $formattedEta = sprintf(
                        (string) __('%s to %s'),
                        $earliestDate,
                        $latestDate
                    );
                } else {
                    $formattedEta = sprintf(
                        '%s',
                        $earliestDate
                    );
                }
            }
        }
        return $formattedEta;
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function updateReservedShipment(
        int $orderId,
        string $shipmentId,
        array $shipmentData
    ): void {
        $transaction = $this->apiHelper->updateReservedShipment(
            $shipmentId,
            $shipmentData
        );
        $this->saveTransaction(
            $transaction,
            (string) $orderId,
            $shipmentId
        );
        if (
            $transaction->getResponseStatusCode() !== 200
        ) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected response status-code from server: %1 and body %2',
                    $transaction->getResponseStatusCode(),
                    $transaction->getResponseBody()
                )
            );
        }
        return;
    }

    public function deleteReservedShipment(
        int $orderId,
        string $shipmentId
    ): void {
        $transaction = $this->apiHelper->deleteReservedShipment(
            $shipmentId
        );
        $this->saveTransaction(
            $transaction,
            $orderId,
            $shipmentId
        );
        if (
            $transaction->getResponseStatusCode() === 200
        ) {
            $decoded =
                $transaction->getResponseBodyDecoded();
            if (empty($decoded['success'])) {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Unexpected response body from server, status: %1 and body: %2',
                        $transaction->getResponseStatusCode(),
                        $transaction->getResponseBody()
                    )
                );
            }
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected response status-code from server: %1 and body %2',
                    $transaction->getResponseStatusCode(),
                    $transaction->getResponseBody()
                )
            );
        }
        return;
    }

    public function getServicePointFormattedName(
        array $servicePoint
    ): string {
        if (isset($servicePoint['visitingAddress']['streetNumber'])) {
            return sprintf(
                '%s, %s %s, %s %s',
                $servicePoint['name'],
                $servicePoint['visitingAddress']['streetName'],
                $servicePoint['visitingAddress']['streetNumber'],
                $servicePoint['visitingAddress']['postalCode'],
                $servicePoint['visitingAddress']['city']
            );
        } else {
            return sprintf(
                '%s, %s, %s %s',
                $servicePoint['name'],
                $servicePoint['visitingAddress']['streetName'],
                $servicePoint['visitingAddress']['postalCode'],
                $servicePoint['visitingAddress']['city']
            );
        }
    }

    public function getConfigurationErrors(): array
    {
        $shipperRegionCode = $this->getConfig(
            \Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_REGION_ID
        );
        if (is_numeric($shipperRegionCode)) {
            $shipperRegionCode =
                $this->regionFactory->create()->load($shipperRegionCode)->getCode();
        }

        $originStreet1 = $this->getConfig(
            \Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_ADDRESS1
        );
        $storeInfo = new \Magento\Framework\DataObject(
            (array) $this->getConfig(
                'general/store_information'
            )
        );

        $errors = [];
        if (
            !$this->getConfig(
                'carriers/mspostnord/automation/username'
            )
        ) {
            $errors[] =
                (string) __('Missing Automation User (Automation > User)');
        }

        $admin = $this->authSession->getUser();
        if ($admin) {
            if (!$admin->getFirstname()) {
                $errors[] =
                    (string) __('Missing Admin Firstname');
            }
            if (!$admin->getLastname()) {
                $errors[] =
                    (string) __('Missing Admin Lastname');
            }
        }
        if (!$storeInfo->getName()) {
            $errors[] =
                (string) __('Missing Store Name (General > General > Store Information > Store Name)');
        }
        if (!$storeInfo->getCountryId()) {
            $errors[] =
                (string) __('Missing Country (General > General > Store Information > Country)');
        }
        if (!$storeInfo->getPhone()) {
            $errors[] =
                (string) __('Missing Store Phone (General > General > Store Information > Store Phone Number)');
        }
        if (!$originStreet1) {
            $errors[] =
                (string) __('Missing Origin Street 1 (Sales > Shipping Settings > Origin > Street Address)');
        }
        if (!$shipperRegionCode) {
            $errors[] =
                (string) __('Missing Shipper Region Code (Sales > Shipping Settings > Origin > Region/State)');
        }
        if (!$this->getConfig(\Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_CITY)) {
            $errors[] =
                (string) __('Missing Store City (Sales > Shipping Settings > Origin > City)');
        }
        if (!$this->getConfig(\Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_ZIP)) {
            $errors[] =
                (string) __('Missing Store ZIP/Postal Code (Sales > Shipping Settings > Origin > ZIP/Postal Code)');
        }
        if (!$this->getConfig(\Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_COUNTRY_ID)) {
            $errors[] =
                (string) __('Missing Store Country ID (Sales > Shipping Settings > Origin > Country)');
        }
        if (!$this->getConfig('carriers/mspostnord/process_mode')) {
            $errors[] =
                (string) __('Missing Process Mode');
        }
        if (!$this->getConfig('carriers/mspostnord/api/identifier')) {
            $errors[] =
                (string) __('Missing API Identifier');
        }
        if (!$this->getConfig('carriers/mspostnord/api/key')) {
            $errors[] =
                (string) __('Missing API Key');
        }
        if (!$this->getConfig('carriers/mspostnord/fallback/title')) {
            $errors[] =
                (string) __('Missing Fallback title');
        }
        if (empty($errors)) {
            $validCredentials =
                $this->isValidApiCredentials();
            if ($validCredentials === -1) {
                $errors[] =
                    (string) __('Failed to connect to API');
            } elseif ($validCredentials === 0) {
                $errors[] =
                    (string) __('Invalid API credentials specified');
            }
        }
        if (!$this->hasProperUsage()) {
            $errors[] =
                (string) __('Invalid product license specified');
        }
        return $errors;
    }

    /**
     * @SuppressWarnings(PHPMD.UnusedFormalParameter)
     */
    public function createOrderReservedShipment(
        int $orderId,
        array $shipmentData
    ): array {
        $transaction = $this->apiHelper->reserveShipment(
            $shipmentData
        );
        $this->saveTransaction(
            $transaction,
            (string) $orderId
        );

        if (
            $transaction->getResponseStatusCode() === 201
        ) {
            $decoded =
                $transaction->getResponseBodyDecoded();
            if (
                !empty(
                    $decoded['reservedShipment']['shipmentId']
                )
            ) {
                if (
                    !empty($decoded['reservedShipment']['status'])
                    && $decoded['reservedShipment']['status'] ===
                    Api::SHIPMENT_STATUS_RESERVED
                ) {
                    return $decoded;
                } else {
                    throw new \Magento\Framework\Exception\LocalizedException(
                        __(
                            'Unexpected response body from server, status: %1 and body: %2',
                            $transaction->getResponseStatusCode(),
                            $transaction->getResponseBody()
                        )
                    );
                }
            } else {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Unexpected response body from server, status: %1 and body: %2',
                        $transaction->getResponseStatusCode(),
                        $transaction->getResponseBody()
                    )
                );
            }
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected response status-code from server: %1 and body %2',
                    $transaction->getResponseStatusCode(),
                    $transaction->getResponseBody()
                )
            );
        }
    }

    public function orderShipmentStatus(
        int $orderId,
        string $shipmentId
    ): array {
        return $this->orderShipmentsStatuses(
            [
                [
                    $orderId,
                    $shipmentId
                ]
            ]
        );
    }

    /**
     * @SuppressWarnings(PHPMD.UnusedFormalParameter)
     */
    public function printShipment(
        int $orderId,
        string $shipmentId
    ): string {
        $transaction = $this->apiHelper->printShipment(
            $shipmentId
        );
        $this->saveTransaction(
            $transaction,
            (string) $orderId,
            $shipmentId
        );
        if (
            $transaction->getResponseStatusCode() === 200
        ) {
            $body =
                $transaction->getResponseBody();
            if (!empty($body)) {
                return base64_decode($body);
            } else {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Unexpected response body from server, status: %1 and body: %2',
                        $transaction->getResponseStatusCode(),
                        $transaction->getResponseBody()
                    )
                );
            }
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected response status-code from server: %1 and body %2',
                    $transaction->getResponseStatusCode(),
                    $transaction->getResponseBody()
                )
            );
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function orderShipmentsStatuses(
        array $ordersAndShipments
    ): array {
        $shipmentIds = [];
        $orderIds = [];
        $shipmentIdToKey = [];
        if (empty($ordersAndShipments)) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Invalid arguments to refresh shipments! Contact developers!'
                )
            );
        }
        foreach ($ordersAndShipments as $orderAndShipment) {
            if (
                count($orderAndShipment) !== 2
                || !is_int($orderAndShipment[0])
                || !is_string($orderAndShipment[1])
            ) {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Invalid arguments to refresh shipments! Contact developers!'
                    )
                );
            }
            $orderIds[] =
                $orderAndShipment[0];
            $shipmentIdToKey[$orderAndShipment[1]] =
                count($shipmentIds);
            $shipmentIds[] =
                $orderAndShipment[1];
        }
        $transaction = $this->apiHelper->getShipments(
            $shipmentIds
        );
        $this->saveTransaction(
            $transaction,
            implode(
                ',',
                $orderIds
            ),
            implode(
                ',',
                $shipmentIds
            )
        );
        if (
            $transaction->getResponseStatusCode() === 200
        ) {
            $decoded =
                $transaction->getResponseBodyDecoded();
            if (
                isset($decoded['shipments'])
                && is_array($decoded['shipments'])
            ) {
                foreach ($decoded['shipments'] as $shipment) {
                    if (
                        empty($shipment['shipmentId'])
                        || !isset($shipmentIdToKey[$shipment['shipmentId']])
                        || empty($shipment['status'])
                        || ($shipment['status'] !== Api::SHIPMENT_STATUS_RESERVED
                            && $shipment['status'] !== Api::SHIPMENT_STATUS_FAILED
                            && $shipment['status'] !== Api::SHIPMENT_STATUS_BOOKED
                            && $shipment['status'] !== Api::SHIPMENT_STATUS_ONGOING
                        )
                    ) {
                        throw new \Magento\Framework\Exception\LocalizedException(
                            __(
                                'Unexpected response body of shipments from server, status: %1 and body: %2',
                                $transaction->getResponseStatusCode(),
                                $transaction->getResponseBody()
                            )
                        );
                    }
                }
                return $decoded;
            } else {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Unexpected response body from server, status: %1 and body: %2',
                        $transaction->getResponseStatusCode(),
                        $transaction->getResponseBody()
                    )
                );
            }
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected response status-code from server: %1 and body %2',
                    $transaction->getResponseStatusCode(),
                    $transaction->getResponseBody()
                )
            );
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function shipmentStatus(
        string $shipmentId
    ): array {
        $transaction = $this->apiHelper->getShipments(
            [$shipmentId]
        );

        $this->saveTransaction(
            $transaction,
            null,
            $shipmentId
        );
        if (
            $transaction->getResponseStatusCode() === 200
        ) {
            $decoded =
                $transaction->getResponseBodyDecoded();
            if (
                isset($decoded['shipments'])
                && is_array($decoded['shipments'])
            ) {
                foreach ($decoded['shipments'] as $shipment) {
                    if (
                        empty($shipment['shipmentId'])
                        || empty($shipment['status'])
                        || ($shipment['status'] !== Api::SHIPMENT_STATUS_RESERVED
                            && $shipment['status'] !== Api::SHIPMENT_STATUS_FAILED
                            && $shipment['status'] !== Api::SHIPMENT_STATUS_BOOKED
                            && $shipment['status'] !== Api::SHIPMENT_STATUS_ONGOING
                        )
                    ) {
                        throw new \Magento\Framework\Exception\LocalizedException(
                            __(
                                'Unexpected response body of shipments from server, status: %1 and body: %2',
                                $transaction->getResponseStatusCode(),
                                $transaction->getResponseBody()
                            )
                        );
                    }
                }
                return $decoded;
            } else {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Unexpected response body from server, status: %1 and body: %2',
                        $transaction->getResponseStatusCode(),
                        $transaction->getResponseBody()
                    )
                );
            }
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected response status-code from server: %1 and body %2',
                    $transaction->getResponseStatusCode(),
                    $transaction->getResponseBody()
                )
            );
        }
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     * @SuppressWarnings(PHPMD.UnusedFormalParameter)
     * phpcs:disable Generic.CodeAnalysis.UnusedFunctionParameter
     */
    public function createOrderBookedShipment(
        int $orderId,
        string $shipmentId
    ): array {
        $transaction = $this->apiHelper->bookShipments(
            [$shipmentId]
        );
        $this->saveTransaction(
            $transaction,
            (string) $orderId,
            $shipmentId
        );

        if (
            $transaction->getResponseStatusCode() === 201
        ) {
            $decoded =
                $transaction->getResponseBodyDecoded();
            if (
                isset($decoded['shipments'])
                && is_array($decoded['shipments'])
            ) {
                foreach ($decoded['shipments'] as $shipment) {
                    if (
                        empty($shipment['shipmentId'])
                        || empty($shipment['status'])
                        || ($shipment['status'] !==
                            Api::SHIPMENT_STATUS_BOOKED
                            && $shipment['status'] !==
                            Api::SHIPMENT_STATUS_ONGOING)
                    ) {
                        throw new \Magento\Framework\Exception\LocalizedException(
                            __(
                                'Unexpected response body of shipments from server, status: %1 and body: %2',
                                $transaction->getResponseStatusCode(),
                                $transaction->getResponseBody()
                            )
                        );
                    }
                }
                return $decoded;
            } else {
                throw new \Magento\Framework\Exception\LocalizedException(
                    __(
                        'Unexpected response body from server, status: %1 and body: %2',
                        $transaction->getResponseStatusCode(),
                        $transaction->getResponseBody()
                    )
                );
            }
        } else {
            throw new \Magento\Framework\Exception\LocalizedException(
                __(
                    'Unexpected response status-code from server: %1 and body %2',
                    $transaction->getResponseStatusCode(),
                    $transaction->getResponseBody()
                )
            );
        }
    }

    public function getConfig(
        string $path,
        string $storeId = null
    ) {
        try {
            $value = $this->scopeConfig->getValue(
                $path,
                \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
                $storeId ?? $this->storeManager->getStore()->getId()
            );

            return $value;
        } catch (\Exception $e) {
            return null;
        }
    }

    public function getAbout(): string
    {
        return '<div><ul style="list-style: circle inside none;">'
            . '<li><strong>' . __('URL') . ':</strong> <a href="' . self::ABOUT_URL . '">'
            . self::ABOUT_URL . '</a></li><li><strong>' . __('Phone') . ':</strong> '
            . '<a href="tel:' . self::ABOUT_PHONE . '">' . self::ABOUT_PHONE . '</a></li>'
            . '<li>' . $this->getSupportEmailLink() . '</li></ul>'
            . '<p style="text-align: left; margin: 20px 0 10px 0;"><a href="'
            . self::ABOUT_URL . '" target="_blank">'
            . '<img alt="M2 PostNord Ship Connect" style="height: 150px; width: auto;" src="' . self::ABOUT_LOGO
            . '" /></a></p></div>';
    }

    public function getSupportEmailLink(): string
    {
        return sprintf(
            '<a href="mailto:%s?subject=%s&body=%s">%s</a>',
            self::ABOUT_EMAIL,
            $this->getMailSubject(),
            $this->getMailBody(),
            (string) __('Click here to create support ticket')
        );
    }

    public function getMailSubject(): string
    {
        return rawurlencode(
            (string) __(
                'I need help to get started with my plugin'
            )
        );
    }

    /**
     * phpcs:disable Generic.Files.LineLength.TooLong
     */
    public function getMailBody(): string
    {
        return rawurlencode(
            (string) __(
                "Describe what you need help with here:"
            ) . "\n\n\n"
                . sprintf(
                    "\n\n---\n%s\nPHP-Version: %s\nPlugin-Version: %s\nMagento2-Version: %s\nPlugin-Name: %s\nDomain: %s",
                    (string) __(
                        'My Environment:'
                    ),
                    \phpversion(),
                    $this->getExtensionVersion(),
                    $this->getMagentoVersion(),
                    'm2-postnord',
                    \parse_url(
                        $this->storeManager->getStore()->getBaseUrl(
                            \Magento\Framework\UrlInterface::URL_TYPE_DIRECT_LINK
                        ),
                        \PHP_URL_HOST
                    )
                )
        );
    }

    public function getExtensionVersion(): string
    {
        static $version;
        if ($version === null) {
            $filename =
                $this->moduleDir->getDir('Mediastrategi_PostNord') . '/VERSION';
            try {
                if (
                    $fileVersion = trim(file_get_contents($filename))
                ) {
                    $version = $fileVersion;
                } else {
                    $version = '-';
                }
            } catch (\Exception $e) {
                return (string) __(
                    'Failed to get extension version!'
                );
            }
        }

        return $version;
    }

    public function getMagentoVersion(): string
    {
        return $this->productMetadata->getVersion();
    }

    public function getExtensionMetaData(): array
    {
        static $metaData = null;
        if ($metaData === null) {
            $metaData = [];
            try {
                if (
                    $encodedContents = \file_get_contents(
                        'https://license.oktagon.se/metadata/m2-postnord.json'
                    )
                ) {
                    $decodedContents = json_decode(
                        $encodedContents,
                        true
                    );
                    if (
                        $decodedContents
                    ) {
                        $metaData =
                            $decodedContents;
                    }
                }
            } catch (\Throwable $e) {
                $metaData['error'] = (string) __(
                    'Failed to get extension metadata!'
                );
            }
        }
        return $metaData;
    }

    public function getChangelog(): array
    {
        $changelog = [];
        if (
            $metaData = $this->getExtensionMetaData()
        ) {
            if (!empty($metaData['changelog'])) {
                $changelog =
                    $metaData['changelog'];
            }
        }
        return $changelog;
    }

    public function getLatestVersion(): string
    {
        $version = (string) __(
            'Failed to get latest extension version!'
        );
        if (
            $metaData = $this->getExtensionMetaData()
        ) {
            if (!empty($metaData['latest_version'])) {
                $version =
                    $metaData['latest_version'];
            }
        }
        return $version;
    }

    public function getLicense(): array
    {
        static $license = null;
        if ($license === null) {
            $license = [
                'domain' => \parse_url(
                    $this->storeManager->getStore()->getBaseUrl(
                        \Magento\Framework\UrlInterface::URL_TYPE_DIRECT_LINK
                    ),
                    \PHP_URL_HOST
                ),
                'password' =>
                    $this->getConfig('carriers/mspostnord/license') ?? '',
            ];
        }
        return $license;
    }

    public function hasProperUsage(): bool
    {
        static $properUsage = null;
        // Use memory cache here
        if ($properUsage === null) {
            $licenseCredentials =
                $this->getLicense();
            // Don't poll license server when lacking credentials
            if (
                empty($licenseCredentials['domain'])
                || empty($licenseCredentials['password'])
            ) {
                $properUsage = false;
                return false;
            }

            // Use file-system cache here
            $license = $this->cacheHelper->loadLicenseValidationResponse(
                $licenseCredentials
            );
            if ($license) {
                $properUsage =
                    !empty($license) && !empty($license['valid']);
            } else {
                $ch = curl_init();
                curl_setopt(
                    $ch,
                    CURLOPT_URL,
                    'https://license.oktagon.se/validate/'
                );
                curl_setopt(
                    $ch,
                    CURLOPT_POST,
                    true
                );
                curl_setopt(
                    $ch,
                    CURLOPT_POSTFIELDS,
                    sprintf(
                        'domain=%s&subscriptionId=%s&product=m2-postnord',
                        urlencode($licenseCredentials['domain']),
                        urlencode($licenseCredentials['password'])
                    )
                );
                curl_setopt(
                    $ch,
                    CURLOPT_FOLLOWLOCATION,
                    true
                );
                curl_setopt(
                    $ch,
                    CURLOPT_MAXREDIRS,
                    10
                );
                curl_setopt(
                    $ch,
                    CURLOPT_ENCODING,
                    ''
                );
                // Setup timeout
                curl_setopt(
                    $ch,
                    CURLOPT_CONNECTTIMEOUT,
                    9000
                );
                curl_setopt(
                    $ch,
                    CURLOPT_TIMEOUT,
                    10000
                );
                curl_setopt(
                    $ch,
                    CURLOPT_RETURNTRANSFER,
                    true
                );

                $responseBody = '';
                $success = false;
                $responseStatusCode = 0;

                try {
                    $responseBody = curl_exec($ch);
                    $responseStatusCode = curl_getinfo(
                        $ch,
                        CURLINFO_HTTP_CODE
                    );
                    $curlErrorMessage = curl_error($ch);
                    curl_close($ch);
                    if (!empty($curlErrorMessage)) {
                        $this->log(
                            sprintf(
                                'License cURL error: %s',
                                $curlErrorMessage
                            )
                        );
                    } else {
                        $success = true;
                    }
                } catch (\Exception $e) {
                    $this->log(
                        sprintf(
                            'License cURL exception: %s',
                            $e->getMessage()
                        )
                    );
                }
                if ($success) {
                    if (
                        $responseStatusCode === 200
                        || $responseStatusCode === 400
                    ) {
                        if (!empty($responseBody)) {
                            $decodedJson =
                                array();
                            try {
                                $decodedJson = json_decode(
                                    $responseBody,
                                    true
                                );
                            } catch (\Exception $e) {
                                $decodedJson = array();
                            }
                            if ($decodedJson) {
                                $license = $decodedJson;
                                $this->cacheHelper->saveLicenseValidationResponse(
                                    $licenseCredentials,
                                    $license
                                );
                                $properUsage =
                                    !empty($license) && !empty($license['valid']);
                            } else {
                                // Invalid JSON means valid license but uncached
                                $properUsage = true;
                            }
                        } else {
                            // Empty response means valid license but uncached
                            $properUsage = true;
                        }
                    } else {
                        // Other errors means invalid license but uncached
                        $properUsage = false;
                    }
                } else {
                    // Network errors means valid license but uncached
                    $properUsage = true;
                }
            }
        }
        return $properUsage;
    }

    /**
     * Magento 2 uses a combination of language and location in locale,
     * PostNord only needs the language part of it.
     */
    public function getLanguageCode(): string
    {
        $locale = '';

        $countryLocale = '';
        if ($this->storeInterface->getStore()) {
            $countryLocale = $this->storeInterface->getStore()->getLocaleCode();
        }
        if (
            !$countryLocale
            && $this->localeResolver
        ) {
            $countryLocale = $this->localeResolver->getLocale();
        }
        if (
            $countryLocale
            && ($explode = explode('_', $countryLocale))
            && !empty(reset($explode))
        ) {
            $locale = reset($explode);
        }

        $forcedLanguage =
            $this->getConfig('carriers/mspostnord/fallback_language');
        if (!$forcedLanguage) {
            $forcedLanguage = '';
        }

        // Forced language setting
        if (
            $forcedLanguage
            && $locale != 'sv'
            && $locale != 'en'
        ) {
            $locale = $forcedLanguage;
        }

        // PostNord API only supports Swedish and English at the moment
        if (
            $locale !== 'sv'
            && $locale !== 'en'
        ) {
            $locale = 'en';
        }

        return $locale;
    }

    public function isEnabledForCheckout(): bool
    {
        return $this->getConfig('carriers/mspostnord/active')
            && $this->hasAccount();
    }

    public function hasAccount(): bool
    {
        return !empty($this->getConfig('carriers/mspostnord/api/identifier'))
            && !empty($this->getConfig('carriers/mspostnord/api/key'));
    }

    public function isDebug(): bool
    {
        return ($this->getConfig('carriers/mspostnord/is_debug')
            == true);
    }

    public function log(string $message): void
    {
        $this->_logger->debug($message);
    }

    private function saveTransaction(
        \Mediastrategi\PostNordAPI\Transaction $transaction,
        string $orderIds = null,
        string $shipmentIds = null,
        string $quoteId = null
    ): ?int {
        return $this->apiTransactionHelper->add(
            $this->getTransactionDataFromTransaction(
                $transaction,
                $orderIds,
                $shipmentIds,
                $quoteId
            )
        );
    }

    /**
     * Gets Transaction object from Query and has an argument if it should be saved to the database
     *
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function getTransactionForQuery(
        array $query,
        string $orderIds = null,
        string $shipmentIds = null,
        string $quoteId = null
    ): ?\Mediastrategi\PostNordAPI\Transaction {
        $transaction = $this->apiHelper->getShippingOptionsForQuery(
            $query
        );

        if ($transaction) {
            $this->saveTransaction(
                $transaction,
                $orderIds,
                $shipmentIds,
                $quoteId
            );
        }

        return $transaction;
    }

    public function getTransactionDataFromTransaction(
        \Mediastrategi\PostNordAPI\Transaction $transaction,
        string $orderIds = null,
        string $shipmentIds = null,
        string $quoteId = null
    ): array {
        return [
            TransactionInterface::ORDER_IDS =>
                $orderIds,
            TransactionInterface::QUOTE_ID =>
                $quoteId,
            TransactionInterface::REQUEST_BODY =>
                $transaction->getRequestBody(),
            TransactionInterface::REQUEST_HEADERS =>
                json_encode($transaction->getRequestHeaders()),
            TransactionInterface::REQUEST_METHOD =>
                $transaction->getRequestMethod(),
            TransactionInterface::REQUEST_URI =>
                $transaction->getRequestUri(),
            TransactionInterface::RESPONSE_BODY =>
                $transaction->getResponseBody(),
            TransactionInterface::RESPONSE_HEADERS =>
                json_encode($transaction->getResponseHeaders()),
            TransactionInterface::RESPONSE_STATUS_CODE =>
                $transaction->getResponseStatusCode(),
            TransactionInterface::SHIPMENT_IDS =>
                $shipmentIds,
        ];
    }

    public function getStatusText(string $status): string
    {
        switch ($status) {
            case \Mediastrategi\PostNord\Helper\Api::SHIPMENT_STATUS_BOOKED:
                return (string) __('Booked');
            case \Mediastrategi\PostNord\Helper\Api::SHIPMENT_STATUS_FAILED:
                return (string) __('Failed');
            case \Mediastrategi\PostNord\Helper\Api::SHIPMENT_STATUS_ONGOING:
                return (string) __('Ongoing');
            case \Mediastrategi\PostNord\Helper\Api::SHIPMENT_STATUS_RESERVED:
                return (string) __('Reserved');
            default:
                return $status;
        }
    }
}
