<?php

declare(strict_types=1);

namespace Mediastrategi\PostNordAPI;

class Connection
{
    const STATUS_ONGOING = 'ongoing';
    const STATUS_BOOKED = 'booked';
    const STATUS_FAILED = 'failed';

    /**
     * @var string
     */
    private $baseUri = '';

    /**
     * @var string
     */
    private $identifier = '';

    /**
     * @var string
     */
    private $key = '';

    /**
     * @var string
     */
    private $labelSize = '';

    /**
     * @var string
     */
    private $paperSize = '';

    /**
     * @var string
     */
    private $token = '';

    /**
     * @throws Exceptions\InvalidArgument
     * @throws Exceptions\InvalidEnvironment
     */
    public function __construct(
        array $configuration
    ) {
        if (!self::isValidConfiguration($configuration)) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid configuration for connection: "%s"!',
                    self::exportValue($configuration)
                )
            );
        }
        if (!self::isValidEnvironment()) {
            throw new Exceptions\InvalidEnvironment(
                sprintf(
                    'Missing cURL or SHA256 features in PHP'
                )
            );
        }
        $this->baseUri =
            $configuration['baseUri'];
        $this->identifier =
            $configuration['identifier'];
        $this->key =
            $configuration['key'];
        if (!empty($configuration['token'])) {
            $this->token =
                $configuration['token'];
        }
        $this->labelSize =
            $configuration['labelSize'];
        $this->paperSize =
            $configuration['paperSize'];
    }

    public function getToken(): string
    {
        return $this->token;
    }

    public static function isValidEnvironment(): bool
    {
        return function_exists('curl_init')
            && function_exists('curl_setopt')
            && function_exists('curl_getinfo')
            && function_exists('curl_close')
            && in_array('sha256', hash_algos());
    }

    public static function isValidRequestBody(
        array $requestBody
    ): bool {
        return self::isValidArray($requestBody);
    }

    public static function isValidConfiguration(
        array $configuration
    ): bool {
        if (
            !self::isValidNonEmptyAssociativeArray(
                $configuration
            )
        ) {
            return false;
        }
        if (
            !isset($configuration['baseUri'])
            || !self::isValidNonEmptyString(
                $configuration['baseUri']
            )
        ) {
            return false;
        }
        if (
            !isset($configuration['identifier'])
            || !self::isValidNonEmptyString(
                $configuration['identifier']
            )
        ) {
            return false;
        }
        if (
            !isset($configuration['key'])
            || !self::isValidNonEmptyString(
                $configuration['key']
            )
        ) {
            return false;
        }
        if (
            isset($configuration['token'])
            && !self::isValidNonEmptyString(
                $configuration['token']
            )
        ) {
            return false;
        }
        if (
            !isset($configuration['labelSize'])
            || !self::isValidLabelSize(
                $configuration['labelSize']
            )
        ) {
            return false;
        }
        if (
            !isset($configuration['paperSize'])
            || !self::isValidPaperSize(
                $configuration['paperSize']
            )
        ) {
            return false;
        }
        return true;
    }

    /**
     * Fetch shipment options that the store owner
     * created in the PostNord Portal.
     *
     * @throws Exceptions\InvalidArgument
     */
    public function postShippingOptionsQuery(
        array $query
    ): Transaction {
        if (
            !self::isValidShippingOptionsQuery(
                $query
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid query: %s!',
                    self::exportValue($query)
                )
            );
        }
        $requestBody = $query;
        return $this->authorizedRequest(
            '/shipping-options/query',
            $requestBody,
            'POST'
        );
    }

    /**
     * Create a shipment reservation based on the given shipping
     * option and other parameters.
     *
     * @throws Exceptions\InvalidArgument
     */
    public function postReservedShipments(
        array $requestBody
    ): Transaction {
        if (
            !self::isValidReservedShipmentsRequestBody(
                $requestBody
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid reserved shipments request body: %s!',
                    self::exportValue($requestBody)
                )
            );
        }
        return $this->authorizedRequest(
            '/reserved-shipments',
            $requestBody,
            'POST'
        );
    }

    /**
     * Update details of a specific shipment reservation.
     *
     * @throws Exceptions\InvalidArgument
     */
    public function putReservedShipments(
        string $reservationId,
        array $requestBody
    ): Transaction {
        if (
            !self::isValidReservationId(
                $reservationId
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid reserved shipment id: %s!',
                    self::exportValue($reservationId)
                )
            );
        }
        if (
            !self::isValidReservedShipmentsRequestBody(
                $requestBody
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid reserved shipments request body: %s!',
                    self::exportValue($requestBody)
                )
            );
        }
        return $this->authorizedRequest(
            sprintf(
                '/reserved-shipments/%s',
                $reservationId
            ),
            $requestBody,
            'PUT'
        );
    }

    /**
     * Delete a shipment reservation.
     *
     * @throws Exceptions\InvalidArgument
     */
    public function deleteReservedShipments(
        string $reservationId
    ): Transaction {
        if (!self::isValidReservationId($reservationId)) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid reservation id: %s!',
                    self::exportValue($reservationId)
                )
            );
        }
        return $this->authorizedRequest(
            sprintf(
                '/reserved-shipments/%s',
                $reservationId
            ),
            [],
            'DELETE'
        );
    }

    /**
     * @throws Exceptions\InvalidArgument
     */
    public function postReservedShipmentsBook(
        array $shipmentsList
    ): Transaction {
        if (!self::isValidShipmentsList($shipmentsList)) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid reserved shipments book shipments list: %s!',
                    self::exportValue($shipmentsList)
                )
            );
        }
        return $this->authorizedRequest(
            '/reserved-shipments/book',
            [
                'shipments' => $shipmentsList,
            ],
            'POST'
        );
    }

    /**
     * @suppress PhanTypeInvalidDimOffset
     */
    public static function isValidReservedShipmentsRequestBody(
        array $requestBody
    ): bool {
        if (
            !self::isValidNonEmptyAssociativeArray(
                $requestBody
            )
        ) {
            return false;
        }
        if (
            !isset($requestBody['selectedShippingOption'])
            || !self::isValidShippingOption(
                $requestBody['selectedShippingOption']
            )
        ) {
            return false;
        }
        if (
            !isset($requestBody['currency'])
            || !is_string($requestBody['currency'])
        ) {
            return false;
        }
        if (
            !isset($requestBody['totalPrice'])
            || !self::isValidInt($requestBody['totalPrice'])
        ) {
            return false;
        }
        if (
            !isset($requestBody['weightInGrams'])
            || !self::isValidInt($requestBody['weightInGrams'])
        ) {
            return false;
        }
        if (
            !isset($requestBody['recipient'])
            || !self::isValidRecipient(
                $requestBody['recipient']
            )
            || empty($requestBody['recipient']['address1'])
            || !is_string($requestBody['recipient']['address1'])
            || empty($requestBody['recipient']['city'])
            || !is_string($requestBody['recipient']['city'])
            || empty($requestBody['recipient']['name'])
            || !is_string($requestBody['recipient']['name'])
        ) {
            return false;
        }
        return true;
    }

    public static function isValidShippingOption(
        array $shippingOption = null
    ): bool {
        if (
            !self::isValidNonEmptyAssociativeArray(
                $shippingOption
            )
        ) {
            return false;
        }
        if (
            !isset($shippingOption['shipmentOptionId'])
            || !self::isValidShipmentOptionId(
                $shippingOption['shipmentOptionId']
            )
        ) {
            return false;
        }
        return true;
    }

    public static function isValidShipmentsList(
        array $shipmentsList = null
    ): bool {
        return self::isValidNumericalArrayOfNonEmptyStrings(
            $shipmentsList
        );
    }

    public static function isValidShipmentOptionId(
        string $shipmentOptionId = null
    ): bool {
        return self::isValidNonEmptyString($shipmentOptionId);
    }

    public static function isValidStatuses(
        string $statuses = null
    ): bool {
        if (
            !self::isValidNonEmptyString(
                $statuses
            )
        ) {
            return false;
        }
        if (
            stripos(
                $statuses,
                ','
            ) !== false
        ) {
            $explode = explode(
                ',',
                $statuses
            );
            foreach ($explode as $status) {
                if (
                    !self::isValidStatus(
                        $status
                    )
                ) {
                    return false;
                }
            }
        } else {
            if (
                !self::isValidStatus(
                    $statuses
                )
            ) {
                return false;
            }
        }
        return true;
    }

    public static function isValidShipmentIds(
        string $shipmentIds = null
    ): bool {
        if (
            !self::isValidNonEmptyString(
                $shipmentIds
            )
        ) {
            return false;
        }
        if (
            stripos(
                $shipmentIds,
                ','
            ) !== false
        ) {
            $explode = explode(
                ',',
                $shipmentIds
            );
            foreach ($explode as $shipmentId) {
                if (
                    !self::isValidShipmentId(
                        $shipmentId
                    )
                ) {
                    return false;
                }
            }
        } else {
            if (
                !self::isValidShipmentId(
                    $shipmentIds
                )
            ) {
                return false;
            }
        }
        return true;
    }

    public static function isValidShipmentId(
        string $shipmentId = null
    ): bool {
        return self::isValidNonEmptyString($shipmentId);
    }

    /**
     * Fetch a list of shipments all shipments
     * booked or reserved the last 14 days.
     */
    public function getShipmentsLast14Days(): Transaction
    {
        return $this->getShipments([]);
    }

    /**
     * Fetch a list of shipments. Query parameters
     * can be used to narrow down the result by using a
     * list of statuses, a date interval or even specific
     * shipment ids. If no parameters are sent, the response
     * will contain all shipments booked or reserved
     * the last 14 days.
     */
    private function getShipments(
        array $args
    ): Transaction {
        return $this->authorizedRequest(
            '/shipments',
            $args,
            'GET'
        );
    }

    /**
     * @throws Exceptions\InvalidArgument
     */
    public function getShipmentsByShipmentIds(
        string $shipmentIds
    ): Transaction {
        if (
            !self::isValidShipmentIds(
                $shipmentIds
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid shipment-ids: %s!',
                    self::exportValue($shipmentIds)
                )
            );
        }
        return $this->getShipments(
            [
                'shipment-ids' =>
                    $shipmentIds,
            ]
        );
    }

    /**
     * @throws Exceptions\InvalidArgument
     */
    public function getShipmentsByStatuses(
        string $statuses
    ): Transaction {
        if (
            !self::isValidStatuses(
                $statuses
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid statuses: %s!',
                    self::exportValue($statuses)
                )
            );
        }
        return $this->getShipments(
            [
                'statuses' => $statuses,
            ]
        );
    }

    /**
     * @throws Exceptions\InvalidArgument
     */
    public function getShipmentsByUnixtime(
        int $from = null,
        int $to = null
    ): Transaction {
        if (
            empty($from)
            && empty($to)
        ) {
            throw new Exceptions\InvalidArgument(
                'Missing from or to unixtime!'
            );
        }
        if (
            !empty($from)
            && !self::isValidInt(
                $from
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid unixtime from: %s!',
                    self::exportValue($from)
                )
            );
        }
        if (
            !empty($to)
            && !self::isValidInt(
                $to
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid unixtime to: %s!',
                    self::exportValue($to)
                )
            );
        }
        $args = [];
        if (!empty($from)) {
            $args['from'] = self::getIso8601DateFromUnixtime(
                $from
            );
        }
        if (!empty($to)) {
            $args['to'] = self::getIso8601DateFromUnixtime(
                $to
            );
        }
        return $this->getShipments(
            $args
        );
    }

    /**
     * Print a list of booked shipment ids
     *
     * @throws Exceptions\InvalidArgument
     */
    public function postShipmentsBulkprint(
        array $shipmentIds
    ): Transaction {
        if (
            !self::isValidNonEmptyNumericalArray(
                $shipmentIds
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid shipment ids: %s!',
                    self::exportValue($shipmentIds)
                )
            );
        }
        return $this->authorizedRequest(
            sprintf(
                '/shipments/bulkprint?paperSize=%s&labelSize=%s',
                $this->paperSize,
                $this->labelSize
            ),
            [
                'ids' => $shipmentIds,
            ],
            'POST'
        );
    }

    /**
     * Print details of a booked shipment.
     *
     * @throws Exceptions\InvalidArgument
     */
    public function postShipments(
        string $bookingId
    ): Transaction {
        if (
            !self::isValidShipmentId(
                $bookingId
            )
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid booking id: %s!',
                    self::exportValue($bookingId)
                )
            );
        }
        return $this->authorizedRequest(
            sprintf(
                '/shipments/%s?paperSize=%s&labelSize=%s',
                $bookingId,
                $this->paperSize,
                $this->labelSize
            ),
            [],
            'POST'
        );
    }

    /**
     * Fetch details of a booked shipment.
     *
     * @throws Exceptions\InvalidArgument
     */
    public function getShipment(
        string $shipmentId
    ): Transaction {
        if (!self::isValidShipmentId($shipmentId)) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid shipment id: %s!',
                    self::exportValue($shipmentId)
                )
            );
        }
        return $this->authorizedRequest(
            sprintf(
                '/shipments/%s',
                $shipmentId
            ),
            [],
            'GET'
        );
    }

    public static function isValidShippingOptionsQuery(
        array $query
    ): bool {
        if (
            !self::isValidNonEmptyAssociativeArray(
                $query
            )
        ) {
            return false;
        }
        if (
            !isset($query['currency'])
            || !self::isValidNonEmptyString($query['currency'])
        ) {
            return false;
        }
        if (
            !isset($query['language'])
            || !self::isValidNonEmptyString($query['language'])
        ) {
            return false;
        }
        if (
            !isset($query['totalPrice'])
            || !self::isValidInt($query['totalPrice'])
        ) {
            return false;
        }
        if (
            !isset($query['weightInGrams'])
            || !self::isValidInt($query['weightInGrams'])
        ) {
            return false;
        }
        if (
            !isset($query['recipient'])
            || !self::isValidRecipient($query['recipient'])
        ) {
            return false;
        }
        return true;
    }

    public static function isValidRecipient(
        array $recipient = null
    ): bool {
        if (
            !self::isValidNonEmptyAssociativeArray($recipient)
            || !isset($recipient['countryCode'])
            || !self::isValidNonEmptyString($recipient['countryCode'])
            || !isset($recipient['postalCode'])
            || !self::isValidNonEmptyString($recipient['postalCode'])
            || !isset($recipient['type'])
            || !self::isValidNonEmptyString($recipient['type'])
        ) {
            return false;
        }
        return true;
    }

    private function hasToken(): bool
    {
        return !empty($this->token);
    }

    private function authorizedRequest(
        string $requestUri,
        array $requestBody = [],
        string $requestMethod = 'GET'
    ): Transaction {
        if (!$this->hasToken()) {
            $this->refreshToken();
        }
        $transaction = $this->request(
            $requestUri,
            $requestBody,
            $requestMethod
        );
        if (
            $transaction->getResponseStatusCode()
            === 401
        ) {
            $this->refreshToken();
            $transaction = $this->request(
                $requestUri,
                $requestBody,
                $requestMethod
            );
        }
        return $transaction;
    }

    /**
     * @throws Exceptions\Curl
     * @throws Exceptions\General
     * @throws Exceptions\InvalidResponse
     * @throws Exceptions\General
     */
    private function refreshToken(): void
    {
        $this->token = '';
        $transaction =
            $this->postAuth();
        switch (
            $transaction->getResponseStatusCode()
        ) {
            case 200:
                $decoded =
                    $transaction->getResponseBodyDecoded();
                if (empty($decoded['token'])) {
                    throw new Exceptions\InvalidResponse(
                        sprintf(
                            'Missing token in response: %s!',
                            $transaction->getResponseBody()
                        )
                    );
                }
                $this->token =
                    $decoded['token'];
                break;
            case 400:
                $decoded =
                    $transaction->getResponseBodyDecoded();
                throw new Exceptions\General(
                    sprintf(
                        "Validation of your request didn't pass. '
                        . The response should have additional info. Message: %s",
                        (!empty($decoded['message'])
                            ? $decoded['message']
                            : '')
                    )
                );
            case 401:
                $decoded =
                    $transaction->getResponseBodyDecoded();
                throw new Exceptions\General(
                    sprintf(
                        "Invalid credentials or expired auth token. Message: %s",
                        (!empty($decoded['message'])
                            ? $decoded['message']
                            : '')
                    )
                );
            default:
                throw new Exceptions\InvalidResponse(
                    sprintf(
                        'Unexpected response status code: %s, body: %s!',
                        $transaction->getResponseStatusCode(),
                        $transaction->getResponseBody()
                    )
                );
        }
        return;
    }

    /**
     * @throws Exceptions\InvalidArgument
     * @throws Exceptions\General
     * @throws Exceptions\Curl
     * @suppress PhanTypeMismatchArgumentNullable
     */
    private function request(
        string $requestUri,
        array $requestBody = [],
        string $requestMethod = 'GET'
    ): Transaction {
        $requestHeaders = [
            'x-bap-key' => 'ost-ecom'
        ];
        $responseHeaders = [];

        // Argument validation
        if (!self::isValidRequestUri($requestUri)) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid request URI "%s"!',
                    self::exportValue($requestUri)
                )
            );
        }
        if (!self::isValidRequestBody($requestBody)) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid request body "%s"!',
                    self::exportValue($requestBody)
                )
            );
        }
        if (!self::isValidRequestMethod($requestMethod)) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid request method "%s"!',
                    self::exportValue($requestMethod)
                )
            );
        }

        $requestUri =
            $this->baseUri . $requestUri;
        if ($requestMethod === 'GET') {
            if (!empty($requestBody)) {
                $requestUri .= '?';
                $getIndex = 0;
                foreach ($requestBody as $key => $value) {
                    if ($getIndex) {
                        $requestUri .= '&';
                    }
                    $decodedKey = urlencode($key);

                    // urlencode actually encodes booleans as integers, API expects actual booleans
                    if (self::isBool($value)) {
                        $decodedValue = ($value ? 'true' : 'false');
                    } else {
                        $decodedValue = urlencode($value);
                    }
                    $requestUri .= $decodedKey . '=' . $decodedValue;
                    $getIndex++;
                }
            }
            $requestBody = '';
        }

        // Setup cURL
        $ch = curl_init();
        curl_setopt(
            $ch,
            CURLOPT_URL,
            $requestUri
        );

        // Setup timeout
        curl_setopt(
            $ch,
            CURLOPT_CONNECTTIMEOUT,
            9000
        );
        curl_setopt(
            $ch,
            CURLOPT_TIMEOUT,
            10000
        );

        if ($this->hasToken()) {
            $requestHeaders['authorization'] = sprintf(
                'Bearer %s',
                $this->getToken()
            );
        }

        if ($requestMethod === 'PUT') {
            curl_setopt(
                $ch,
                CURLOPT_CUSTOMREQUEST,
                'PUT'
            );
            if (!empty($requestBody)) {
                $requestHeaders['content-type'] =
                    'application/json';
                $requestBody =
                    json_encode($requestBody);
                curl_setopt(
                    $ch,
                    CURLOPT_POSTFIELDS,
                    $requestBody
                );
            } else {
                $requestBody = '';
            }
        }

        if ($requestMethod === 'DELETE') {
            curl_setopt(
                $ch,
                CURLOPT_CUSTOMREQUEST,
                'DELETE'
            );
            if (!empty($requestBody)) {
                $requestHeaders['content-type'] =
                    'application/json';
                $requestBody =
                    json_encode($requestBody);
                curl_setopt(
                    $ch,
                    CURLOPT_POSTFIELDS,
                    $requestBody
                );
            } else {
                $requestBody = '';
            }
        }

        if ($requestMethod === 'POST') {
            curl_setopt(
                $ch,
                CURLOPT_POST,
                true
            );
            if (!empty($requestBody)) {
                $requestHeaders['content-type'] =
                    'application/json';
                $requestBody =
                    json_encode($requestBody);
                curl_setopt(
                    $ch,
                    CURLOPT_POSTFIELDS,
                    $requestBody
                );
            } else {
                $requestBody = '';
            }
        }

        curl_setopt(
            $ch,
            CURLOPT_RETURNTRANSFER,
            true
        );

        if (!empty($requestHeaders)) {
            $requestHeaders2 = [];
            foreach ($requestHeaders as $key => $value) {
                $requestHeaders2[] = sprintf(
                    '%s: %s',
                    $key,
                    $value
                );
            }
            curl_setopt(
                $ch,
                CURLOPT_HTTPHEADER,
                $requestHeaders2
            );
        }

        curl_setopt(
            $ch,
            CURLOPT_HEADERFUNCTION,
            function (
                $ch,
                $header
            ) use (&$responseHeaders) {
                $len = strlen($header);
                $header = explode(
                    ':',
                    $header,
                    2
                );
                if (count($header) < 2) {
                    return $len;
                }

                $responseHeaders[strtolower(trim($header[0]))][] =
                    trim($header[1]);
                return $len;
            }
        );

        // Perform request
        $responseBody = '';
        try {
            $responseBody = curl_exec($ch);
            $responseStatusCode = curl_getinfo(
                $ch,
                CURLINFO_HTTP_CODE
            );
            $curlErrorMessage = curl_error($ch);
            curl_close($ch);
            if (!empty($curlErrorMessage)) {
                throw new Exceptions\Curl(
                    $curlErrorMessage
                );
            }
        } catch (\Exception $e) {
            throw new Exceptions\General(
                $e->getMessage()
            );
        }

        return new Transaction(
            $requestMethod,
            $requestUri,
            $requestHeaders,
            $requestBody,
            $responseStatusCode,
            $responseHeaders,
            $responseBody
        );
    }

    /**
     * This auth token is a jwt and will be used in the
     * Authorization header for subsequent requests:
     * Authorization: Bearer {token}. The client provides
     * the store's identifier and a random generated
     * string called nonce. The client appends the
     * store's key to the nonce and runs the Sha256
     * algorithm on that value. This creates a digest
     * value that is also provided in this request's body.
     * digest = sha256(nonce+key)
     *
     * @throws \Exception
     */
    private function postAuth(): Transaction
    {
        $nonce = uniqid();
        $digest = hash(
            'sha256',
            $nonce . $this->key
        );
        $requestBody = [
            'identifier' =>
                $this->identifier,
            'nonce' =>
                $nonce,
            'digest' =>
                $digest,
        ];
        return $this->request(
            '/auth',
            $requestBody,
            'POST'
        );
    }

    public static function isValidReservationId(
        string $reservationId = null
    ): bool {
        return self::isValidNonEmptyString($reservationId);
    }

    private static $validRequestMethods = [
        'DELETE' => true,
        'GET' => true,
        'POST' => true,
        'PUT' => true
    ];

    public static function isValidStatus(
        $status
    ): bool {
        return (self::isValidNonEmptyString($status)
            && ($status === self::STATUS_BOOKED
                || $status === self::STATUS_ONGOING
                || $status === self::STATUS_FAILED));
    }

    public static function exportValue(
        $value
    ): string {
        if (
            is_array($value)
            || is_object($value)
        ) {
            return print_r(
                $value,
                true
            );
        }
        return (string) $value;
    }

    public static function isValidNumericalArray(
        $array
    ): bool {
        if (!self::isValidArray($array)) {
            return false;
        }
        foreach (array_keys($array) as $key) {
            if (!self::isValidInt($key)) {
                return false;
            }
        }
        return true;
    }

    public static function isBool(
        $bool
    ): bool {
        return is_bool($bool);
    }

    public static function isValidNonEmptyAssociativeArray(
        $array
    ): bool {
        return
            self::isValidAssociativeArray($array)
            && !empty($array);
    }

    public static function isValidNonEmptyNumericalArray(
        $array
    ): bool {
        return
            self::isValidNumericalArray($array)
            && !empty($array);
    }

    public static function isValidAssociativeArray(
        $array
    ): bool {
        if (!self::isValidArray($array)) {
            return false;
        }
        if (empty($array)) {
            return true;
        }
        foreach (array_keys($array) as $key) {
            if (self::isValidString($key)) {
                return true;
            }
        }
        return false;
    }

    public static function isValidArray(
        $array = null
    ): bool {
        return is_array($array);
    }

    public static function isValidString(
        $string
    ): bool {
        return is_string($string);
    }

    public static function isValidNonEmptyString(
        $string
    ): bool {
        return self::isValidString($string)
            && !empty($string);
    }

    public static function isValidInt(
        $int
    ): bool {
        return isset($int)
            && is_int($int);
    }

    public static function isValidNonEmptyInt(
        $int
    ): bool {
        return self::isValidInt($int)
            && !empty($int);
    }

    public static function isValidNumericalArrayOfNonEmptyStrings(
        $list
    ): bool {
        if (
            empty($list)
            || !is_array($list)
        ) {
            return false;
        }
        foreach ($list as $key => $value) {
            if (!self::isValidInt($key)) {
                return false;
            }
            if (
                !self::isValidNonEmptyString(
                    $value
                )
            ) {
                return false;
            }
        }
        return true;
    }

    public static function isValidRequestUri(
        $uri
    ): bool {
        if (
            empty($uri)
            || !is_string($uri)
        ) {
            return false;
        }
        return true;
    }

    public static function isValidRequestMethod(
        $method
    ): bool {
        if (
            empty($method)
            || !is_string($method)
            || !isset(self::$validRequestMethods[$method])
        ) {
            return false;
        }
        return true;
    }

    public static function getIso8601DateFromUnixtime(
        int $unixTime
    ): string {
        return date(
            "Y-m-d\TH:i:s\Z",
            $unixTime
        );
    }

    public static function isValidDate(
        $date
    ): bool {
        if (
            !self::isValidNonEmptyString(
                $date
            )
        ) {
            return false;
        }
        return preg_match(
            '/^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]+Z$/',
            $date
        ) === 1;
    }

    public static function isValidPaperSize($paperSize): bool
    {
        return !empty($paperSize)
            && is_string($paperSize)
            && ($paperSize === 'A5'
                || $paperSize === 'A4'
                || $paperSize === 'LETTER'
                || $paperSize === 'LABEL');
    }

    public static function isValidLabelSize($labelSize): bool
    {
        return !empty($labelSize)
            && is_string($labelSize)
            && ($labelSize === 'standard'
                || $labelSize === 'small');
    }
}
