<?php

declare(strict_types=1);

namespace Mediastrategi\PostNordAPI;

class Helper
{
    const STATUS_ONGOING = 'ongoing';
    const STATUS_BOOKED = 'booked';
    const STATUS_FAILED = 'failed';

    private static $validRequestMethods = [
        'DELETE' => true,
        'GET' => true,
        'POST' => true,
        'PUT' => true
    ];

    /**
     * @throws Exceptions\InvalidArgument
     */
    public static function getTrackingUri(
        string $trackingId
    ): string {
        if (
            !self::isValidNonEmptyString($trackingId)
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid tracking-id: %s!',
                    self::exportValue($trackingId)
                )
            );
        }
        return sprintf(
            'https://portal.postnord.com/tracktrace/#/myShipments/details/v2/%s',
            $trackingId
        );
    }

    /**
     * @throws Exceptions\InvalidArgument
     */
    public static function getTestTrackingUri(
        string $trackingId
    ): string {
        if (
            !self::isValidNonEmptyString($trackingId)
        ) {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid tracking-id: %s!',
                    self::exportValue($trackingId)
                )
            );
        }
        return sprintf(
            'https://atportal.postnord.com/trackingWESXHFKG/details/%s',
            $trackingId
        );
    }

    public static function isValidStatus(
        $status
    ): bool {
        return (self::isValidNonEmptyString($status)
            && ($status === self::STATUS_BOOKED
                || $status === self::STATUS_ONGOING
                || $status === self::STATUS_FAILED));
    }

    public static function exportValue(
        $value
    ): string {
        if (
            is_array($value)
            || is_object($value)
        ) {
            return print_r(
                $value,
                true
            );
        }
        return $value;
    }

    public static function isValidNumericalArray(
        $array
    ): bool {
        if (!self::isValidArray($array)) {
            return false;
        }
        foreach (array_keys($array) as $key) {
            if (!self::isValidInt($key)) {
                return false;
            }
        }
        return true;
    }

    public static function isBool(
        $bool
    ): bool {
        return is_bool($bool);
    }

    public static function isValidNonEmptyAssociativeArray(
        $array
    ): bool {
        return
            self::isValidAssociativeArray($array)
            && !empty($array);
    }

    public static function isValidNonEmptyNumericalArray(
        $array
    ): bool {
        return
            self::isValidNumericalArray($array)
            && !empty($array);
    }

    public static function isValidAssociativeArray(
        $array
    ): bool {
        if (!self::isValidArray($array)) {
            return false;
        }
        if (empty($array)) {
            return true;
        }
        foreach (array_keys($array) as $key) {
            if (self::isValidString($key)) {
                return true;
            }
        }
        return false;
    }

    public static function isValidArray(
        $array = null
    ): bool {
        return is_array($array);
    }

    public static function isValidString(
        $string
    ): bool {
        return is_string($string);
    }

    public static function isValidNonEmptyString(
        $string
    ): bool {
        return self::isValidString($string)
            && !empty($string);
    }

    public static function isValidInt(
        $int
    ): bool {
        return isset($int)
            && is_int($int);
    }

    public static function isValidNonEmptyInt(
        $int
    ): bool {
        return self::isValidInt($int)
            && !empty($int);
    }

    public static function isValidNumericalArrayOfNonEmptyStrings(
        $list
    ): bool {
        if (
            empty($list)
            || !is_array($list)
        ) {
            return false;
        }
        foreach ($list as $key => $value) {
            if (!self::isValidInt($key)) {
                return false;
            }
            if (
                !self::isValidNonEmptyString(
                    $value
                )
            ) {
                return false;
            }
        }
        return true;
    }

    public static function isValidRequestUri(
        $uri
    ): bool {
        if (
            empty($uri)
            || !is_string($uri)
        ) {
            return false;
        }
        return true;
    }

    public static function isValidRequestMethod(
        $method
    ): bool {
        if (
            empty($method)
            || !is_string($method)
            || !isset(self::$validRequestMethods[$method])
        ) {
            return false;
        }
        return true;
    }

    public static function getIso8601DateFromUnixtime(
        int $unixTime
    ): string {
        return date(
            "Y-m-d\TH:i:s\Z",
            $unixTime
        );
    }

    public static function isValidDate(
        $date
    ): bool {
        if (
            !self::isValidNonEmptyString(
                $date
            )
        ) {
            return false;
        }
        return preg_match(
            '/^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]+Z$/',
            $date
        ) === 1;
    }
}
