<?php

declare(strict_types=1);

namespace Mediastrategi\PostNordAPI;

class Transaction
{
    const STATUS_ONGOING = 'ongoing';
    const STATUS_BOOKED = 'booked';
    const STATUS_FAILED = 'failed';

    private static $validRequestMethods = [
        'DELETE' => true,
        'GET' => true,
        'POST' => true,
        'PUT' => true
    ];

    /**
     * @var string
     */
    private $requestMethod = '';

    /**
     * @var string
     */
    private $requestUri = '';

    /**
     * Numerical indexed array.
     *
     * @var array
     */
    private $requestHeaders = [];

    /**
     * @var string
     */
    private $requestBody = '';

    /**
     * @var array|null
     */
    private $requestBodyDecoded = null;

    /**
     * @var int
     */
    private $responseStatusCode = 0;

    /**
     * Associative indexed array.
     *
     * @var array
     */
    private $responseHeaders = [];

    /**
     * @var string
     */
    private $responseBody = '';

    /**
     * @var array|null
     */
    private $responseBodyDecoded = null;

    /**
     * @throws Exceptions\InvalidArgument
     */
    public function __construct(
        string $requestMethod,
        string $requestUri,
        array $requestHeaders,
        string $requestBody,
        int $responseStatusCode,
        array $responseHeaders,
        string $responseBody
    ) {
        if (
            self::isValidRequestMethod(
                $requestMethod
            )
        ) {
            $this->requestMethod =
                $requestMethod;
        } else {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid request method "%s"!',
                    self::exportValue($requestMethod)
                )
            );
        }
        if (
            self::isValidRequestUri(
                $requestUri
            )
        ) {
            $this->requestUri =
                $requestUri;
        } else {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid request uri "%s"!',
                    self::exportValue($requestUri)
                )
            );
        }
        if (
            self::isValidRequestHeaders(
                $requestHeaders
            )
        ) {
            $this->requestHeaders =
                $requestHeaders;
        } else {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid request headers "%s"!',
                    self::exportValue($requestHeaders)
                )
            );
        }
        if (
            self::isValidRequestBody(
                $requestBody
            )
        ) {
            $this->requestBody =
                $requestBody;
        } else {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid request body "%s"!',
                    self::exportValue($requestBody)
                )
            );
        }
        if (
            self::isValidResponseStatusCode(
                $responseStatusCode
            )
        ) {
            $this->responseStatusCode =
                $responseStatusCode;
        } else {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid response status code "%s"!',
                    self::exportValue($responseStatusCode)
                )
            );
        }
        if (
            self::isValidResponseHeaders(
                $responseHeaders
            )
        ) {
            $this->responseHeaders =
                $responseHeaders;
        } else {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid response headers "%s"!',
                    self::exportValue($responseHeaders)
                )
            );
        }
        if (
            self::isValidResponseBody(
                $responseBody
            )
        ) {
            $this->responseBody =
                $responseBody;
        } else {
            throw new Exceptions\InvalidArgument(
                sprintf(
                    'Invalid response body "%s"!',
                    self::exportValue($responseBody)
                )
            );
        }
    }

    /**
     * @param mixed $requestBody
     */
    public static function isValidRequestBody(
        $requestBody = null
    ): bool {
        return self::isValidString($requestBody);
    }

    /**
     * @param mixed $requestHeaders
     */
    public static function isValidRequestHeaders(
        $requestHeaders = null
    ): bool {
        return self::isValidAssociativeArray(
            $requestHeaders
        );
    }

    /**
     * @param mixed $responseHeaders
     */
    public static function isValidResponseHeaders(
        $responseHeaders = null
    ): bool {
        return self::isValidAssociativeArray(
            $responseHeaders
        );
    }

    /**
     * @param mixed $responseBody
     */
    public static function isValidResponseBody(
        $responseBody = null
    ): bool {
        return self::isValidString($responseBody);
    }

    /**
     * @param mixed $responseStatusCode
     */
    public static function isValidResponseStatusCode(
        $responseStatusCode = null
    ): bool {
        return self::isValidNonEmptyInt(
            $responseStatusCode
        );
    }

    public function getRequestHeaders(): array
    {
        return $this->requestHeaders;
    }

    public function getRequestMethod(): string
    {
        return $this->requestMethod;
    }

    public function getRequestUri(): string
    {
        return $this->requestUri;
    }

    public function getRequestBody(): string
    {
        return $this->requestBody;
    }

    public function getResponseStatusCode(): int
    {
        return $this->responseStatusCode;
    }

    public function getResponseBody(): string
    {
        return $this->responseBody;
    }

    public function getResponseBodyDecoded(): array
    {
        if (!isset($this->responseBodyDecoded)) {
            try {
                $this->responseBodyDecoded = json_decode(
                    $this->responseBody,
                    true
                );
            } catch (\Exception $e) {
                $this->responseBodyDecoded = [];
            }
        }
        return (array) $this->responseBodyDecoded;
    }

    public function getRequestBodyDecoded(): array
    {
        if (!isset($this->requestBodyDecoded)) {
            try {
                $this->requestBodyDecoded = json_decode(
                    $this->requestBody,
                    true
                );
            } catch (\Exception $e) {
                $this->requestBodyDecoded = [];
            }
        }
        return (array) $this->requestBodyDecoded;
    }

    public function getResponseHeaders(): array
    {
        return $this->responseHeaders;
    }

    public function __toString(): string
    {
        return print_r(
            $this,
            true
        );
    }

    public static function isValidStatus(
        $status
    ): bool {
        return (self::isValidNonEmptyString($status)
            && ($status === self::STATUS_BOOKED
                || $status === self::STATUS_ONGOING
                || $status === self::STATUS_FAILED));
    }

    public static function exportValue(
        $value
    ): string {
        if (
            is_array($value)
            || is_object($value)
        ) {
            return print_r(
                $value,
                true
            );
        }
        return (string) $value;
    }

    public static function isValidNumericalArray(
        $array
    ): bool {
        if (!self::isValidArray($array)) {
            return false;
        }
        foreach (array_keys($array) as $key) {
            if (!self::isValidInt($key)) {
                return false;
            }
        }
        return true;
    }

    public static function isBool(
        $bool
    ): bool {
        return is_bool($bool);
    }

    public static function isValidNonEmptyAssociativeArray(
        $array
    ): bool {
        return
            self::isValidAssociativeArray($array)
            && !empty($array);
    }

    public static function isValidNonEmptyNumericalArray(
        $array
    ): bool {
        return
            self::isValidNumericalArray($array)
            && !empty($array);
    }

    public static function isValidAssociativeArray(
        $array
    ): bool {
        if (!self::isValidArray($array)) {
            return false;
        }
        if (empty($array)) {
            return true;
        }
        foreach (array_keys($array) as $key) {
            if (self::isValidString($key)) {
                return true;
            }
        }
        return false;
    }

    public static function isValidArray(
        $array = null
    ): bool {
        return is_array($array);
    }

    public static function isValidString(
        $string
    ): bool {
        return is_string($string);
    }

    public static function isValidNonEmptyString(
        $string
    ): bool {
        return self::isValidString($string)
            && !empty($string);
    }

    public static function isValidInt(
        $int
    ): bool {
        return isset($int)
            && is_int($int);
    }

    public static function isValidNonEmptyInt(
        $int
    ): bool {
        return self::isValidInt($int)
            && !empty($int);
    }

    public static function isValidNumericalArrayOfNonEmptyStrings(
        $list
    ): bool {
        if (
            empty($list)
            || !is_array($list)
        ) {
            return false;
        }
        foreach ($list as $key => $value) {
            if (!self::isValidInt($key)) {
                return false;
            }
            if (
                !self::isValidNonEmptyString(
                    $value
                )
            ) {
                return false;
            }
        }
        return true;
    }

    public static function isValidRequestUri(
        $uri
    ): bool {
        if (
            empty($uri)
            || !is_string($uri)
        ) {
            return false;
        }
        return true;
    }

    public static function isValidRequestMethod(
        $method
    ): bool {
        if (
            empty($method)
            || !is_string($method)
            || !isset(self::$validRequestMethods[$method])
        ) {
            return false;
        }
        return true;
    }

    public static function getIso8601DateFromUnixtime(
        int $unixTime
    ): string {
        return date(
            "Y-m-d\TH:i:s\Z",
            $unixTime
        );
    }

    public static function isValidDate(
        $date
    ): bool {
        if (
            !self::isValidNonEmptyString(
                $date
            )
        ) {
            return false;
        }
        return preg_match(
            '/^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]+Z$/',
            $date
        ) === 1;
    }
}
