<?php

declare(strict_types=1);

namespace Mediastrategi\PostNord\ViewModel\Adminhtml;

class ApiTester implements \Magento\Framework\View\Element\Block\ArgumentInterface
{
    /**
     * @var string
     */
    private $arguments = '';

    /**
     * @var string
     */
    private $log = '';

    /**
     * @var array
     */
    private $options = [];

    /**
     * @var \Magento\Framework\UrlInterface
     */
    private $urlBuilder;
    /**
     * @var \Mediastrategi\PostNord\Helper\Meta
     */
    private $metaHelper;
    /**
     * @var \Magento\Framework\App\RequestInterface
     */
    private $request;
    /**
     * @var \Magento\Store\Model\StoreManagerInterface
     */
    private $storeManager;

    public function __construct(
        \Magento\Framework\UrlInterface $urlBuilder,
        \Mediastrategi\PostNord\Helper\Meta $metaHelper,
        \Magento\Framework\App\RequestInterface $request,
        \Magento\Store\Model\StoreManagerInterface $storeManager
    ) {
        $this->urlBuilder   = $urlBuilder;
        $this->metaHelper   = $metaHelper;
        $this->request      = $request;
        $this->storeManager = $storeManager;
    }

    /**
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \JsonException
     * @throws \Exception
     */
    public function startTest(): void
    {
        $this->log((string) __('Test started'));
        $successCount = 0;
        $errorsCount  = 0;

        $arguments = [
            'currency'      => $this->storeManager->getStore()->getBaseCurrency()->getCurrencyCode(),
            'language'      => $this->metaHelper->getLanguageCode(),
            'recipient'     => [
                'countryCode' => $this->metaHelper->getConfig(
                    \Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_COUNTRY_ID
                ),
                'postalCode'  => str_replace(
                    ' ',
                    '',
                    $this->metaHelper->getConfig(
                        \Magento\Sales\Model\Order\Shipment::XML_PATH_STORE_ZIP
                    )
                ),
                'type'        => 'personal',
            ],
            'totalPrice'    => random_int(500, 20000),
            'weightInGrams' => 2000,
        ];

        $post = (array) $this->request->getPost();
        if (!empty($post['arguments'])) {
            try {
                $decodedArguments = json_decode($post['arguments'], true, 512, JSON_THROW_ON_ERROR);
            } catch (\JsonException $e) {
                $decodedArguments = [];
            }
            if (
                !empty($decodedArguments)
                && is_array($decodedArguments)
            ) {
                $arguments = $decodedArguments;
            } else {
                $this->log(
                    (string) __(
                        'ERROR: Specified JSON was invalid: "%1"',
                        $post['arguments']
                    )
                );
            }
        }
        $this->arguments = json_encode($arguments, JSON_THROW_ON_ERROR);
        $this->log(
            (string) __(
                'Arguments for REST request: %1',
                var_export($arguments, true)
            )
        );

        try {
            $transaction = $this->metaHelper->getTransactionForQuery(
                $arguments
            );

            if ($transaction) {
                $decoded       = $transaction->getResponseBodyDecoded();
                $this->options = (!empty($decoded['shipmentOptions'])
                    ? $decoded['shipmentOptions'] : []);
                $this->log(
                    (string) __(
                        'SUCCESS: POST /shipping-options/query, shipmentOptions: %1',
                        var_export(
                            $this->options,
                            true
                        )
                    )
                );
                $this->log(
                    (string) __(
                        "HTTP Request Method: %1",
                        $transaction->getRequestMethod()
                    )
                );
                $this->log(
                    (string) __(
                        "HTTP Request URI: %1",
                        $transaction->getRequestUri()
                    )
                );
                $this->log(
                    (string) __(
                        "HTTP Request Headers:\n%1",
                        var_export($transaction->getRequestHeaders(), true)
                    )
                );
                $this->log(
                    (string) __(
                        "HTTP Request Body:\n%1",
                        $transaction->getRequestBody()
                    )
                );
                $this->log(
                    (string) __(
                        "HTTP Response Status Code: %1",
                        $transaction->getResponseStatusCode()
                    )
                );
                $this->log(
                    (string) __(
                        "HTTP Response Body:\n%1",
                        $transaction->getResponseBody()
                    )
                );
                $successCount++;
            } else {
                $this->log(
                    (string) __(
                        'Request failed, please verify your credentials.'
                    )
                );
            }
        } catch (\Exception $e) {
            $this->log(
                (string) __(
                    'ERROR: Request Shipping Options, error: %1, arguments: %2',
                    $e->getMessage(),
                    var_export($arguments, true)
                )
            );
            $errorsCount++;
        }

        $this->log(
            (string) __(
                'Test finished, successes: %1, errors: %2',
                $successCount,
                $errorsCount
            )
        );
    }

    public function getOptions(): array
    {
        return $this->options;
    }

    public function getLog(): string
    {
        return $this->log;
    }

    public function getArguments(): string
    {
        return $this->arguments;
    }

    private function log(string $message): void
    {
        $line = date('Y-m-d H:i:s') . ' | ' . $message . "\n";
        $this->log .= $line;
    }

    public function getActionUrl(): string
    {
        return $this->urlBuilder->getUrl('mspostnord/misc/test');
    }
}
